/************************************************************************
          Copyright 1988, 1991 by Carnegie Mellon University

                          All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted, provided
that the above copyright notice appear in all copies and that both that
copyright notice and this permission notice appear in supporting
documentation, and that the name of Carnegie Mellon University not be used
in advertising or publicity pertaining to distribution of the software
without specific, written prior permission.

CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
IN NO EVENT SHALL CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS
ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
************************************************************************/

/*
 * Bootstrap Protocol (BOOTP).  RFC951 and RFC1395.
 *
 * $Id: bootp.h $
 *
 *
 * This file specifies the "implementation-independent" BOOTP protocol
 * information which is common to both client and server.
 *
 */

#include "bptypes.h"	/* for int32, u_int32 */

#define BP_CHADDR_LEN	 16
#define BP_SNAME_LEN	 64
#define BP_FILE_LEN	128
#define BP_VEND_LEN	 64
#define BP_MINPKTSZ	300	/* to check sizeof(struct bootp) */

struct bootp {
    byte             bp_op;			/* packet opcode type */
    byte             bp_htype;			/* hardware addr type */
    byte             bp_hlen;			/* hardware addr length */
    byte             bp_hops;			/* gateway hops */
    u_int32_t        bp_xid;			/* transaction ID */
    unsigned short   bp_secs;			/* seconds since boot began */
    unsigned short   bp_flags;			/* RFC1532 broadcast, etc. */
    struct in_addr   bp_ciaddr;			/* client IP address */
    struct in_addr   bp_yiaddr;			/* 'your' IP address */
    struct in_addr   bp_siaddr;			/* server IP address */
    struct in_addr   bp_giaddr;			/* gateway IP address */
    byte             bp_chaddr[BP_CHADDR_LEN];	/* client hardware address */
    char	     bp_sname[BP_SNAME_LEN];	/* server host name */
    char	     bp_file[BP_FILE_LEN];	/* boot file name */
    byte             bp_vend[BP_VEND_LEN];	/* vendor-specific area */
    /* note that bp_vend can be longer, extending to end of packet. */
};

/*
 * UDP port numbers, server and client.
 */
#define	IPPORT_BOOTPS		67
#define	IPPORT_BOOTPC		68

#define BOOTREPLY		2
#define BOOTREQUEST		1

/*
 * Hardware types from Assigned Numbers RFC.
 */
#define HTYPE_ETHERNET		  1
#define HTYPE_EXP_ETHERNET	  2
#define HTYPE_AX25		  3
#define HTYPE_PRONET		  4
#define HTYPE_CHAOS		  5
#define HTYPE_IEEE802		  6
#define HTYPE_ARCNET		  7

/*
 * Vendor magic cookie (v_magic) for CMU
 */
#define VM_CMU		"CMU"

/*
 * Vendor magic cookie (v_magic) for RFC1048
 */
#define VM_RFC1048	{ 99, 130, 83, 99 }



/*
 * Tag values used to specify what information is being supplied in
 * the vendor (options) data area of the packet.
 */
/* RFC 1048 */
#define TAG_END			((byte) 255)
#define TAG_PAD			((byte)   0)
#define TAG_SUBNET_MASK		((byte)   1)
#define TAG_TIME_OFFSET		((byte)   2)
#define TAG_GATEWAY		((byte)   3)
#define TAG_TIME_SERVER		((byte)   4)
#define TAG_NAME_SERVER		((byte)   5)
#define TAG_DOMAIN_SERVER	((byte)   6)
#define TAG_LOG_SERVER		((byte)   7)
#define TAG_COOKIE_SERVER	((byte)   8)
#define TAG_LPR_SERVER		((byte)   9)
#define TAG_IMPRESS_SERVER	((byte)  10)
#define TAG_RLP_SERVER		((byte)  11)
#define TAG_HOST_NAME		((byte)  12)
#define TAG_BOOT_SIZE		((byte)  13)
/* RFC 1395 */
#define TAG_DUMP_FILE		((byte)  14)
#define TAG_DOMAIN_NAME		((byte)  15)
#define TAG_SWAP_SERVER		((byte)  16)
#define TAG_ROOT_PATH		((byte)  17)
/* RFC 1497 */
#define TAG_EXTEN_FILE		((byte)  18)
/* RFC 1533 */
#define TAG_NIS_DOMAIN		((byte)  40)
#define TAG_NIS_SERVER		((byte)  41)
#define TAG_NTP_SERVER		((byte)  42)
/* Added for auto/dynamic IP addressing */
#define TAG_DYNAMIC_OPTIONS	((byte) 250)
#define TAG_DYNAMIC_FORCE	((byte) 253)
#define TAG_DYNAMIC_SIZE	((byte) 254)
/* DHCP maximum message size. */
#define TAG_DHCP_IPLEASE	((byte)  51)
#define TAG_DHCP_MSG	   	((byte)  53)
#define TAG_DHCP_SERVERID	((byte)  54)
#define TAG_MAX_MSGSZ		((byte)  57)
#define TAG_DHCP_IPRENEW	((byte)  58)	/* PeP hic facet */
#define TAG_DHCP_IPREBIND	((byte)  59)
/* XXX - Add new tags here */


/*
 * "vendor" data permitted for CMU bootp clients.
 */

struct cmu_vend {
	char		v_magic[4];	/* magic number */
	u_int32_t	v_flags;	/* flags/opcodes, etc. */
	struct in_addr 	v_smask;	/* Subnet mask */
	struct in_addr 	v_dgate;	/* Default gateway */
	struct in_addr	v_dns1, v_dns2; /* Domain name servers */
	struct in_addr	v_ins1, v_ins2; /* IEN-116 name servers */
	struct in_addr	v_ts1, v_ts2;	/* Time servers */
	int32_t		v_unused[6];	/* currently unused */
};


/* v_flags values */
#define VF_SMASK	1	/* Subnet mask field contains valid data */
